	TITLE	"eForth17"

;   C. H. Ting, April 1993

	LIST	LINES=62,COLUMNS=132
	FILL	0

	LIST l=62,c=132     ;62 lines per page, 132 characters per line

;===============================================================
;
;	PIC eForth 1.0, C. H. Ting, Jan. 1993
;	eForth 1.0 by Bill Muench and C. H. Ting, 1990
;	8086 figForth by Thomas Newman, 1981 and Joe smith, 1983
;
;		Dr. C. H. Ting
;		156 14th Avenue
;		San Mateo, CA 94402
;		(415) 571-7639
;
;	4/3/93	Interpreter working
;	4/6/93	Compiler working
;	4/8/93	Constants, variables all tested
;	4/14/93 Update source files, add comments
;
;   WARNING
;	pic17 is a 16 bit cell machine.  Memory is addressed in
;	16 bit cells.  eForth assumes byte addressibility to access
;	character strings.  The resolution is to create an artificial
;	byte addressing space in which byte addresses are 2x the
;	cell addresses.  High order byte has lower, even address
;	and low order byte has higher, odd address.  This is the
;	little-ending order Motorola prefers.
;
;	Most memory addressing words assume cell addresses.  Byte
;	addressing words, however, assume byte addresses.  If you
;	encounter any problem in using this system, most likely
;	you are using the wrong assumption.  Watch the comments
;	in the source listing.	Cell addresses are denoted by a
;	and byte addresses are denoted by b.  The following
;	words are byte addressing words:
;	C@, C!, -TRAILING, CMOVE, TYPE, _TYPE, >CHAR, FILL, PARSE,
;	TAP, kTAP, accept, EXPECT.
;
;	The following words are mixed mode:
;	ALIGNED, COUNT, and PACK$
;
;	YOU HAVE BEEN WARNED!
;
;	4/13/93     Add clrwdt to DOLST.
;

;	Acknowledgements:
;
;	Mike Jedrzejewski of Microchip Technology for technical
;	    support and the EDP17 Evaluation Module
;	Ron Blake for the PIC-Master ICE and Pro-Master programmer
;

;	More information on eForth:
;
;	There are 12 implementations of eForth on many micro-
;	processors, including 8086, Z80, 8051, 68000, 68332,
;	68HC11, 8096/98, 80386 protected mode.	The source
;	code of these implementations are available from
;	Offete Enterprises, 1306 South B St., San Mateo, CA
;	94402, Tel (415) 574-8250, on diskette at $25.00 each.
;
;	The primary documentation on eForth is supplied with the
;	8086 implementation, which is the basic model of all the
;	other implementations.
;
;	All eForth implementations are in the public domain to
;	encourage programmers to learn Forth and make use of it.
;	New implementations are welcome.  Send it to Dr. Ting
;	c/o Offete Enterprises.
;
;	This file is to be assembled by ASM17, v1.6 or later.
;	The assembly and linking commands are:
;
;	    >ASM17 EFORTH17
;	    >LINK17 EFORTH17
;
;	The resulting .HEX file can be used to program EPROMs,
;	or loaded into the Pro-Master In-circuit Emulator for
;	testing.
;
;	eForth17 is coded so that a virgin PIC17C42 chip can
;	execute the code in the microprocessor mode, without
;	having to program the configuration fuses.  For this
;	reason, the colon interpreter DOLST ticks the watchdog
;	timer every time a colon word is executed.  This overhead
;	is very small and worthwhile.
;
;	eForth17 is rather fast for a high level language
;	superposed on an 8-bit microprocessor.	Its
;	speed is comparable to the eForth on a 8086 class
;	microprocessor, because of the RISC-like PIC17 CPU, which
;	executes most instructions in a single machine cycle.
;
;	In eForth17, only the generic eForth word set is
;	implemented.  To make use of the peripheral facilities
;	built in PIC17C42, the user will have to extend the
;	kernal to include words addressing the peripheral
;	registers.  They are conveniently left out as exercises
;	for the students.
;
;===============================================================


;   Virtual Forth Engine Implemented on PIC17C42

;   PIC17 is an 8 bit CPU with 16 bit instructions. The 8 bit
;   ALU really limits its capability to host an efficient Forth
;   Virtual Engine.  It is a challenge to implement a Forth on
;   it with good performance.  One fundamental limitation of PIC17
;   is that there are only two ways to access external program
;   memory: through the PC to fetch the next instruction and
;   through the TBLPTR register pair to read/write memory as a
;   table.  The Forth engine has to read nested lists from the
;   program memory, to fetch and store data from/to the program
;   memory, and manipulate data on two stacks.	These activities
;   are not quite compatible with PIC, which has its humble origin
;   as an 8 bit controller.

;   A redeeming virtue in PIC is the fairly large internal register
;   space, with 232 8 bits registers available to the user.  My
;   strategy is to place the data stack and return stack in this
;   register space, the system dictionary in the ROM space from
;   0 to 3FFF, the user dictionary and variables in the RAM space
;   from 4000 to 7FFF.	The space 8000 to FFFF will be left open
;   to implement a ROM or RAM disk.

;   PIC has two indirectly addressible registers with auto post-
;   increment and auto post-decrement capability.  To build stacks
;   efficiently, you need pre-decrement-post-increment-registers.
;   So, I can only build two half-decent stacks with auto-post-
;   incrementing-manual-decrementing registers.  That's life. Sad.

;   PIC has a 16 level deep stack for subroutine threading.  I am
;   tempted to use it as the return stack in a subroutine threaded
;   Forth engine.  However, there are many serious problems against
;   this design.  The first is the concern that 16 level stack may
;   not be sufficient for arbitrary Forth applications.  The second
;   is that it is very difficult to recover from stack over/under flow.
;   A minor problem is that this design will preclude multitasking.
;   Memory allocation, Microprocessor mode


;   Register allocation

	CBLOCK	0x00
	    SPP 	    ; indirect register to access data stack
	    SP		    ; post-incrementing data stack pointer
	    PCL 	    ; program counter
	    PCH
	    ALUSTA
	    RTCSTA
	    CPUSTA
	    INTSTA
	    RPP 	    ; indirect register to access return stack
	    RP		    ; post-incrmenting return stack pointer
	    W		    ; CPU working register
	    RTCCL
	    RTCCH
	    TBL 	    ; auto-incrementing memory pointer
	    TBH 	    ; TBL-TBH are the Forth IP
	    BSR 	    ; bank select register
	ENDC

	CBLOCK	0x10
	    PORTA
	    DDRB
	    PORTB
	    RCSTA
	    RCREG
	    TXSTA
	    TXREG
	    SPBRG
	ENDC

	CBLOCK	0x10
	    DDRC
	    PORTC
	    DDRD
	    PORTD
	    DDRE
	    PORTE
	    PIR
	    PIE
	ENDC

	CBLOCK	0X10
	    TMR1
	    TMR2
	    TMR3L
	    TMR3H
	    PR1
	    PR2
	    PR3L
	    PR3H
	ENDC

	CBLOCK	0X18	    ; This is the Forth Virtual Machine
	    TL		    ; TL-TH top of data stack
	    TH
	    WPL 	    ; word pointer lower byte
	    WPH 	    ; word pointer high byte
	    XL		    ; scratch
	    XH		    ; scratch
	    YL		    ; scratch
	    YH		    ; scratch
	ENDC

RPCPU	EQU	0x20	    ; return stack in CPU RAM
SPCPU	EQU	0x90	    ; data stack in CPU RAM

ROMCOD	EQU	0x0000	    ; ROM area 0-3FFF
ROMNAM	EQU	0x0C00	    ; bottom of name dictionary
RAM	EQU	0x4000	    ; RAM area 4000-7FFF
UPP	EQU	0x4000	    ; first variable in RAM
HIEXE	EQU	0x4028	    ; goto code area for high ram
USERA	EQU	0x4080	    ; default user area
TIBB	EQU	0x8200	    ; terminal input buffer, 2*0x4100
NEWCOD	EQU	0x4180	    ; for user defined code
NEWNAM	EQU	0x7F00	    ; for user defined names

DISK	EQU	0x8000	    ; solid state disk


;   Macro definitions

PUSH:	MACRO		    ; push TL-TH to data stack
	movfp	TL,SPP	    ; auto-increment SP
	movfp	TH,SPP
	ENDM

POP:	MACRO		    ; pop data stack to TL-TH
	movlw	2	    ; pre-decrementing SP
	subwf	SP,1
	movpf	SPP,TL
	movpf	SPP,TH
	subwf	SP,1	    ; realign SP after popping
	ENDM

SAVE:	MACRO		    ; save TL-TH in XL-XH
	movpf	TL,XL
	movpf	TH,XH
	ENDM

RSTR:	MACRO		    ; restore TL-TH from XL-XH
	movfp	XL,TL
	movfp	XH,TH
	ENDM

NEXT:	MACRO		    ; the address interpreter
	goto	GONXT
	ENDM

;   System EQU's

BASEE	EQU	10


;   Here starts the bootup and interrupt handlers

START:	ORG	0
	GOTO	COLDD
	DW	"eForth17"  ; bury a few text strings in space
INTPORT:ORG	8
	retfie
	DW	"v1.01"
TIMER:	ORG	0x10
	retfie
	DW	"C. H. Ting"
RTPIN:	ORG	0x18
	retfie
	DW	"April 1993"
PERIP:	ORG	0x20
	retfie
	DW	"SVFIG"
INNER:	ORG	0x28

; inner interpreters: NEXT DOLST DOVAR DOCON DOUSE

; These interpreters must be copied to the beginning of every
; 8K code pages so that they can be invoked by GOTO instructions.
; All create-does words must have their own interpreters
; installed in this area.

; In this implementation, the user code dictionary is in
; the 4000-7FFF area.  Code area 28H-7FH is copied to
; 4028H-407FH to support user defined code.

; Defining the inner interpreters for variables, constants,
; and user variables in code deviates from the original
; eForth model.  However, they are examples illustrating
; how users can define their own classes of words like
; the CREATE-DOES> structures in regular Forth systems.

GONXT: tablrd	0,1,XL	    ; read next word in list
	tlrd	1,WPH	    ; get upper byte of execution address
	tlrd	0,WPL	    ; save lower byte of execution address
	movfp	WPH,PCH     ; prepare to jump
	movfp	WPL,PCL     ; copy address to PC and jump to target

DOLST: movfp	TBL,RPP     ; save IP on the return stack
	movfp	TBH,RPP
	movfp	WPH,TBH     ; copy execution address to IP
	movfp	WPL,TBL
	tablrd	0,1,XL	    ; increment IP pointing to address list
	clrwdt		    ; clear watchdog timer
	NEXT		    ; start executing the addresses in list

DOVAR:	PUSH		    ; make room in TOP of data stack
	movpf	WPH,TH	    ; copy execution address to TOP
	movpf	WPL,TL
	clrf	W
	incf	TL,1	    ; increment TOP
	addwfc	TH,1	    ; point to body where variable is
	NEXT

DOCON: PUSH		    ; make room in TOP
	movpf	TBL,XL	    ; save IP in XL-XH
	movpf	TBH,XH
	movpf	WPH,TBH     ; copy execution address into TBL-TBH
	movpf	WPL,TBL
	tablrd	0,1,TL	    ; point TBL-TBH to constant in body
	tablrd	0,0,TL	    ; read constant
	tlrd	0,TL	    ; get constant to TOP
	tlrd	1,TH
	movfp	XL,TBL	    ; restore IP
	movfp	XH,TBH
	NEXT

DOUSE: PUSH		    ; make room in TOP
	movpf	TBL,XL	    ; save IP in XL-XH
	movpf	TBH,XH
	movpf	WPH,TBH     ; copy execution address into TBL-TBH
	movpf	WPL,TBL
	tablrd	0,1,TL	    ; point TBL-TBH to user offset in body
	tablrd	0,0,TL
	tlrd	0,TL	    ; get user offset
	tlrd	1,TH
	movlw	HIGH UPP    ; UP at 0x4000, pointing to 0x4080 user area
	movwf	TBH
	movlw	UPP
	movwf	TBL
	tablrd	0,0,YL	    ; get UP pointer
	tlrd	0,W
	addwf	TL,1	    ; add UP to user offset
	tlrd	1,W
	addwfc	TH,1	    ; address of user variable in TOP
	movfp	XL,TBL	    ; restore IP
	movfp	XH,TBH
	NEXT


; Kernel words in PIC assembly

EXIT:	movlw	2	    ; pop IP off the return stack
	subwf	RP,1
	movpf	RPP,TBL     ; put IP back into TBL-TBH
	movpf	RPP,TBH
	subwf	RP,1
	NEXT		    ; unnest

EXECU:movfp	TH,WPH	    ; copy TOP into PC
	movfp	TL,WPL	    ; save low address in WPL
	POP		    ; pop TOP
	movfp	WPH,PCH
	movfp	WPL,PCL     ; jump to target address

BRAN:	tablrd	0,1,XL	    ; read address in next word
	tlrd	0,TBL	    ; put address in IP
	tlrd	1,TBH
	NEXT		    ; unconditional branch

QBRAN:	movfp	TL,W	    ; TOP=0?
	iorwf	TH,0
	movwf	XL
	POP
	tstfsz	XL
	goto	QBRAN1	    ; TOP not 0, skip next word
	goto	BRAN	    ; TOP is 0. branch to next address
QBRAN1: tablrd	0,1,XL	    ; skip next address and continue on
	NEXT

DONXT: movlw	2	    ; for FOR-NEXT loop
	subwf	RP,1	    ; decrement RP to access loop count
	movpf	RPP,W	    ; is loop count 0?
	iorwf	RPP,0
	tstfsz	W
	goto	NEXT1	    ; count not 0
	movlw	2	    ; count is 0, pop count off return stack
	subwf	RP,1
	tablrd	0,1,XL	    ; skip branch address
	NEXT		    ; exit the loop
NEXT1:	movlw	2	    ; count not 0.
	subwf	RP,1
	decf	RPP,1	    ; decrement loop count
	clrf	W
	subwfb	RPP,1
	goto	BRAN

DOLIT:	PUSH		    ; make room in TOP
	tablrd	0,1,XL	    ; read next word as a literal
	tlrd	0,TL	    ; copy literal to TOP
	tlrd	1,TH
	NEXT		    ; continue

DROP:	POP		    ; pop data stack to TOP
	NEXT

DUPP:	PUSH		    ; push TOP to data stack
	NEXT

OVER:	movlw	2	    ; copy next item on stack to XL-XH
	subwf	SP,1
	movpf	SPP,XL
	movpf	SPP,XH
	PUSH		    ; push TOP to data stack
	RSTR		    ; copy XL-XH to TOP
	NEXT

SWAP:	SAVE		    ; save TOP to XL-XH
	POP		    ; move next to TOP
	movfp	XL,SPP	    ; push XL-XH to data stack
	movfp	XH,SPP
	NEXT

SPSTO:	movfp	TL,SP	    ; set SP 2 bytes above the data stack
	POP		    ; pop the extra 2 bytes into TOP
	NEXT

SPAT:	PUSH		    ; make room in TOP
	clrf	TH
	movpf	SP,TL	    ; get stack pointer in SP
	NEXT

TOR:	movfp	TL,RPP	    ; push TOP on the return stack
	movfp	TH,RPP
	POP		    ; pop data stack
	NEXT

RFROM:	PUSH		    ; make room in TOP
	movlw	2	    ; copy return stack into TOP
	subwf	RP,1
	movpf	RPP,TL
	movpf	RPP,TH
	subwf	RP,1	    ; pop return stack
	NEXT

RAT:	PUSH		    ; make room in TOP
	movlw	2	    ; copy return stack to TOP
	subwf	RP,1
	movpf	RPP,TL
	movpf	RPP,TH
	NEXT		    ; do not pop return stack

RPSTO:	movfp	TL,RP	    ; reset return stack pointer
	POP		    ; discard TOP
	NEXT

RPAT:	PUSH		    ; make room in TOP
	clrf	TH
	movpf	RP,TL	    ; get RP into TOP
	NEXT

ANDD:	movlw	2	    ; prepare to access next on data stack
	subwf	SP,1
	movpf	SPP,W	    ; get low byte in next
	andwf	TL,1	    ; AND with TL
	movpf	SPP,W	    ; get high byte in next
	andwf	TH,1	    ; AND with TH
	movlw	2
	subwf	SP,1	    ; discard next on data stack
	NEXT

ORR:	movlw	2	    ; ditto
	subwf	SP,1
	movpf	SPP,W
	iorwf	TL,1
	movpf	SPP,W
	iorwf	TH,1
	movlw	2
	subwf	SP,1
	NEXT

XORR:	movlw	2	    ; ditto
	subwf	SP,1
	movpf	SPP,W
	xorwf	TL,1
	movpf	SPP,W
	xorwf	TH,1
	movlw	2
	subwf	SP,1
	NEXT

ZLESS:	btfsc	TH,7	    ; test bit 15 in TH, the sign bit
	goto	ZLESS1
	clrf	TH	    ; clear TOP if sign bit is clear
	clrf	TL
	NEXT
ZLESS1: setf	TH	    ; set TOP if sign bit is set
	setf	TL
	NEXT

UPLUS:	movlw	2	    ; access next item on data stack
	subwf	SP,1
	movfp	TL,W
	addwf	SPP,1	    ; add TL to low byte in next
	movfp	TH,W
	addwfc	SPP,1	    ; add TH to high byte in next
	clrf	TH
	clrf	TL
	clrf	W
	addwfc	TL	    ; put carry in TL
	NEXT

AT:	movpf	TBL,XL	    ; save IP
	movpf	TBH,XH
	movfp	TL,TBL	    ; copy cell address into TBL-TBH
	movfp	TH,TBH
	tablrd	0,0,TL	    ; read cell
	tlrd	0,TL	    ; copy cell data to TOP
	tlrd	1,TH
	movfp	XL,TBL	    ; restore IP
	movfp	XH,TBH
	NEXT

STORE:	movpf	TBL,XL	    ; save IP
	movpf	TBH,XH
	movfp	TL,TBL	    ; copy cell address into TBL-TBH
	movfp	TH,TBH
	movlw	2	    ; access next item on data stack
	subwf	SP,1
	tlwt	0,SPP
	tablwt	1,0,SPP     ; write data in next to memory
	subwf	SP,1	    ; discard next
	subwf	SP,1
	movpf	SPP,TL	    ; discard TOP
	movpf	SPP,TH
	subwf	SP,1
	movfp	XL,TBL	    ; restore IP
	movfp	XH,TBH
	NEXT

;   PIC memory is organized in 16 bit cells.
;   Bytes can be accessed with C@ and C! only in the lower 32K region.
;   In a byte address, bits 1-15 is shifted right to form a cell address.
;   Bit 0 selects high or low byte in a cell.
;
;   Since the pic assembler puts strings in a high-first-low-next order,
;   an even byte address accesses the high byte in a cell, and an odd byte
;   address accesses the low byte in a cell.
;   C@ and C! are coded accordingly.

CAT:	movpf	TBL,XL	    ; save IP
	movpf	TBH,XH
	rrcf	TL,0	    ; byte address = address/2
	rrcf	TH,1	    ; odd byte in bits 0-7
	rrcf	TL,1	    ; even byte in bits 8-15
	movfp	TL,TBL
	movfp	TH,TBH
	bcf	TBH,7	    ; clear even-odd bit before fetching
	tablrd	0,0,TL	    ; fetch whole cell
	btfss	TH,7	    ; even byte or odd byte?
	goto	CAT1
	tlrd	0,TL	    ; odd byte is taken from bits 0-7
	goto	CAT2
CAT1:	tlrd	1,TL	    ; even byte from bits 8-15
CAT2:	clrf	TH,1
	movfp	XL,TBL	    ; restore IP
	movfp	XH,TBH
	NEXT

CSTOR:	movpf	TBL,XL	    ; save IP
	movpf	TBH,XH
	rrcf	TL,0	    ; byte address = address/2
	rrcf	TH,1	    ; odd byte in bits 0-7
	rrcf	TL,1	    ; even byte in bits 8-15
	movfp	TL,TBL
	movfp	TH,TBH
	bcf	TBH,7	    ; clear even-odd bit before fetching
	tablrd	0,0,TL	    ; fetch whole cell
	movlw	2
	subwf	SP,1	    ; point to byte to be stored
	btfss	TH,7	    ; even byte or odd byte?
	goto	CSTO1
	tlrd	1,TL
	tlwt	1,TL
	tablwt	0,0,SPP     ; odd byte is stored in bits 0-7
	goto	CSTO2
CSTO1:	tlrd	0,TL
	tlwt	0,TL
	tablwt	1,0,SPP     ; even byte is stored in bits 8-15
CSTO2:	decf	SP,1	    ; discard byte stored
	POP		    ; discard address in TL-TH
	movfp	XL,TBL	    ; restore IP
	movfp	XH,TBH
	NEXT

; MOVE is coded to facilitate copying the user area
; during booting.  This MOVE takes cell addresses and
; cell count.  It is different from CMOVE in the system
; which uses byte addresses and byte count.

MOVE:	movfp	TBL,RPP     ; save IP on return stack
	movfp	TBH,RPP
	movlw	4
	subwf	SP,1
	movpf	SPP,XL	    ; copy source addr to X
	movpf	SPP,XH
	movpf	SPP,YL	    ; copy destination addr to Y
	movpf	SPP,YH
	subwf	SP,1
MOVE1:	tstfsz	TL	    ; test count in TOP
	goto	MOVE1+5     ; not 0, go do MOVE2
	tstfsz	TH
	goto	MOVE1+5
	goto	MOVE1+22    ; TOP=0, exit to MOVE3
MOVE2:	decf	TL,1	    ; decrement count in TOP
	clrf	W
	subwfb	TH,1
	movfp	XL,TBL	    ; get source addr
	movfp	XH,TBH
	tablrd	0,1,W	    ; fetch source data
	tlrd	0,W	    ; read low byte for writing back
	tlrd	1,PCH	    ; sorry. out of usable register
	movpf	TBL,XL	    ; save incremented source addr
	movpf	TBH,XH
	movfp	YL,TBL	    ; get destination addr
	movfp	YH,TBH
	tlwt	1,PCH	    ; write back high byte
	tablwt	0,1,W	    ; store data to destination
	movpf	TBL,YL	    ; save incremented destination addr
	movpf	TBH,YH
	goto	MOVE1	    ; loop until count is zero
MOVE3:	POP		    ; discard count in TOP
	subwf	RP,1	    ; restore IP from return stack
	movpf	RPP,TBL
	movpf	RPP,TBH
	subwf	RP,1
	NEXT


; Hardware dependent code for serial I/O port

STOIO:	movlw	0x20	    ; Async mode, Tx enable, 8 bits
	movwf	TXSTA
	movlw	0x90	    ; Rx enable, 8 bits, serial port enable
	movwf	RCSTA
;	movlw	15	    ; 9600 @ 10 MHz
;	movlw	32	    ; 9600 @ 20 MHz
	movlw	25	    ; 9600 @ 16 MHz
	movwf	SPBRG	    ; serial port baud rate generator
	NEXT

QRX:	PUSH
	clrf	TH
	btfss	RCSTA,1     ; overrun error?
	goto	QRX1
	bcf	RCSTA,4     ; clear CREN
	bsf	RCSTA,4     ; enable receiver again
	goto	QRX3
QRX1:	movlb	1	    ; bank 1
	btfss	PIR,0	    ; test RBFL in PIR
	goto	QRX2	    ; jump to false flag
	movlb	0
	movpf	RCREG,TL    ; get character
	PUSH		    ; push true flag
	setf	TH
	setf	TL
	NEXT
QRX2:	movlb	0	    ; no character available
QRX3:	clrf	TL	    ; place false flag
	NEXT

TXSTO:	movlb	1
TXSTO1: btfss	PIR,1	    ; wait until TSR is empty
	goto	TXSTO1
	movlb	0
	movfp	TL,TXREG    ; send byte
	POP
	NEXT


; Pointer to the user area

UP:	goto	DOCON	    ; variable as a constant in ROM
	DW	UPP


; Load the high level code

;; Version control

VER		EQU	01			;major release version
EXT		EQU	01			;minor extension

;; Constants

COMPO		EQU	0X4000			;lexicon compile only bit
IMEDD		EQU	0X8000			;lexicon immediate bit
MASKK		EQU	0X1F7F			;lexicon bit mask

CALLL		EQU	0XC02D			; goto DOLST
USEEE		EQU	0XC049			; goto DOUSE
VARRR		EQU	0XC034			; goto DOVAR
CONNN		EQU	0XC03C			; goto DOCON

CELLL		EQU	2			;size of a cell
;BASEE		 EQU	 10			 ;default radix
VOCSS		EQU	8			;depth of vocabulary stack

BKSPP		EQU	8			;backspace
LF		EQU	10			;line feed
CRR		EQU	13			;carriage return
ERR		EQU	27			;error escape
TIC		EQU	39			;tick


;; Memory allocation

;EM		 EQU	 0X8000 		 ;top of memory
;			 0X7FF0 		 ;user name dictionary
;			 0X4200 		 ;user code dictionary
;			 0X4100 		 ;terminal input buffer
;			 0X4080 		 ;user variables
;			 0X4028 		 ;user inner interpreters
;			 0X4000 		 ;system variables
;			 0X0FFF 		 ;end of name dictionary
;			 0X0890 		 ;end of code dictionary
;			 0X0028 		 ;code dictionary
;COLDD		 EQU	 0X0000 		 ;cold start vector


;; System and user variables

SZERO:		goto	DOUSE
		DW	4
RZERO:		goto	DOUSE
		DW	5
TQKEY:		goto	DOUSE
		DW	6
TEMIT:		goto	DOUSE
		DW	7
TEXPE:		goto	DOUSE
		DW	8
TTAP:		goto	DOUSE
		DW	9
TECHO:		goto	DOUSE
		DW	10
TPROM:		goto	DOUSE
		DW	11
BASE:		goto	DOUSE
		DW	12
TEMP:		goto	DOUSE
		DW	13
SPAN:		goto	DOUSE
		DW	14
INN:		goto	DOUSE
		DW	15
NTIB:		goto	DOUSE
		DW	16
CSP:		goto	DOUSE
		DW	18
TEVAL:		goto	DOUSE
		DW	19
TNUMB:		goto	DOUSE
		DW	20
HLD:		goto	DOUSE
		DW	21
HANDL:		goto	DOUSE
		DW	22
CNTXT:		goto	DOUSE
		DW	23
CRRNT:		goto	DOUSE
		DW	32
CP:		goto	DOUSE
		DW	34
NP:		goto	DOUSE
		DW	35
LAST:		goto	DOUSE
		DW	36


;; Common functions

;   doVOC	( -- )
;		Run time action of VOCABULARY"s.

DOVOC:  goto    DOLST
		DW	RFROM,CNTXT,STORE,EXIT	;for new vocabularies

;   FORTH	( -- )
;		Make FORTH the context vocabulary.

FORTH:  goto    DOLST
		DW	DOLIT,0X40A5		;voc pointer in RAM
		DW	CNTXT,STORE		;make FORTH the context
		DW	EXIT			;vocabulary

;   ?DUP	( w -- w w | 0 )
;		Dup tos if its is not zero.

QDUP:   goto    DOLST
                        		DW	DUPP
		DW	QBRAN,QDUP+5
		DW	DUPP
QDUP1:		DW	EXIT

;   ROT		( w1 w2 w3 -- w2 w3 w1 )
;		Rot 3rd item to top.

ROT:    goto    DOLST
		DW	TOR,SWAP,RFROM,SWAP,EXIT

;   2DROP	( w w -- )
;		Discard two items on stack.

DDROP:  goto    DOLST
		DW	DROP,DROP,EXIT

;   2DUP	( w1 w2 -- w1 w2 w1 w2 )
;		Duplicate top two items.

DDUP:   goto    DOLST
        		DW	OVER,OVER,EXIT

;   +		( w w -- sum )
;		Add top two items.

PLUS:   goto    DOLST
        		DW	UPLUS,DROP,EXIT

;   D+		( d d -- d )
;		Double addition, as an example using UM+.
;
;		DW	TOR,SWAP,TOR,UPLUS
;		DW	RFROM,RFROM,PLUS,PLUS,EXIT

;   NOT		( w -- w )
;		One's complement of tos.

INVER:  goto    DOLST
    		DW	DOLIT,-1,XORR,EXIT

;   NEGATE	( n -- -n )
;		Two's complement of tos.

NEGAT:  goto    DOLST
    		DW	INVER,DOLIT,1,PLUS,EXIT

;   DNEGATE	( d -- -d )
;		Two's complement of top double.

DNEGA:	goto	DOLST
		DW	INVER,TOR,INVER
		DW	DOLIT,1,UPLUS
		DW	RFROM,PLUS,EXIT

;   -		( n1 n2 -- n1-n2 )
;		Subtraction.

SUBB:   goto    DOLST
		DW	NEGAT,PLUS,EXIT

;   ABS		( n -- n )
;		Return the absolute value of n.

ABSS:   goto    DOLST
		DW	DUPP,ZLESS
		DW	QBRAN,ABSS+6
		DW	NEGAT
ABS1:		DW	EXIT

;   =		( w w -- t )
;		Return true if top two are equal.

EQUAL:  goto    DOLST
		DW	XORR
		DW	QBRAN,EQUAL+7
		DW	DOLIT,0,EXIT		;false flag
EQU1:		DW	DOLIT,-1,EXIT		;true flag

;   U<		( u u -- t )
;		Unsigned compare of top two items.

ULESS:  goto    DOLST
		DW	DDUP,XORR,ZLESS
		DW	QBRAN,ULESS+10
		DW	SWAP,DROP,ZLESS,EXIT
ULES1:		DW	SUBB,ZLESS,EXIT

;   <		( n1 n2 -- t )
;		Signed compare of top two items.

LESS:   goto    DOLST
		DW	DDUP,XORR,ZLESS
		DW	QBRAN,LESS+9
		DW	DROP,ZLESS,EXIT
LESS1:		DW	SUBB,ZLESS,EXIT

;   MAX		( n n -- n )
;		Return the greater of two top stack items.

MAX:    goto    DOLST
		DW	DDUP,LESS
		DW	QBRAN,MAX+6
		DW	SWAP
MAX1:		DW	DROP,EXIT

;   MIN		( n n -- n )
;		Return the smaller of top two stack items.

MIN:    goto    DOLST
		DW	DDUP,SWAP,LESS
		DW	QBRAN,MIN+7
		DW	SWAP
MIN1:		DW	DROP,EXIT

;   WITHIN	( u ul uh -- t )
;		Return true if u is within the range of ul and uh.

WITHI:  goto    DOLST
		DW	OVER,SUBB,TOR			;ul <= u < uh
		DW	SUBB,RFROM,ULESS,EXIT


;; Divide

;   UM/MOD	( udl udh u -- ur uq )
;		Unsigned divide of a double by a single. Return mod and quotient.

UMMOD:  goto    DOLST
		DW	DDUP,ULESS
		DW	QBRAN,UMMOD+43
		DW	NEGAT,DOLIT,15,TOR
UMM1:		DW	TOR,DUPP,UPLUS
		DW	TOR,TOR,DUPP,UPLUS
		DW	RFROM,PLUS,DUPP
		DW	RFROM,RAT,SWAP,TOR
		DW	UPLUS,RFROM,ORR
		DW	QBRAN,UMMOD+36
		DW	TOR,DROP,DOLIT,1,PLUS,RFROM
		DW	BRAN,UMMOD+37
UMM2:		DW	DROP
UMM3:		DW	RFROM
		DW	DONXT,UMM1
		DW	DROP,SWAP,EXIT
UMM4:		DW	DROP,DDROP
		DW	DOLIT,-1,DUPP,EXIT	;overflow, return max

;   M/MOD	( d n -- r q )
;		Signed floored divide of double by single. Return mod and quotient.

MSMOD:  goto    DOLST
		DW	DUPP,ZLESS,DUPP,TOR
		DW	QBRAN,MSMOD+11
		DW	NEGAT,TOR,DNEGA,RFROM
MMOD1:		DW	TOR,DUPP,ZLESS
		DW	QBRAN,MSMOD+18
		DW	RAT,PLUS
MMOD2:		DW	RFROM,UMMOD,RFROM
		DW	QBRAN,MSMOD+26
		DW	SWAP,NEGAT,SWAP
MMOD3:		DW	EXIT

;   /MOD	( n n -- r q )
;		Signed divide. Return mod and quotient.

SLMOD:  goto    DOLST
		DW	OVER,ZLESS,SWAP,MSMOD,EXIT

;   MOD		( n n -- r )
;		Signed divide. Return mod only.

MODD:   goto    DOLST
		DW	SLMOD,DROP,EXIT

;   /		( n n -- q )
;		Signed divide. Return quotient only.

SLASH:  goto    DOLST
		DW	SLMOD,SWAP,DROP,EXIT


;; Multiply

;   UM*		( u u -- ud )
;		Unsigned multiply. Return double product.

UMSTA:  goto    DOLST
		DW	DOLIT,0,SWAP,DOLIT,15,TOR
UMST1:		DW	DUPP,UPLUS,TOR,TOR
		DW	DUPP,UPLUS,RFROM,PLUS,RFROM
		DW	QBRAN,UMSTA+23
		DW	TOR,OVER,UPLUS,RFROM,PLUS
UMST2:		DW	DONXT,UMST1
		DW	ROT,DROP,EXIT

;   *		( n n -- n )
;		Signed multiply. Return single product.

STAR:   goto    DOLST
		DW	UMSTA,DROP,EXIT

;   M*		( n n -- d )
;		Signed multiply. Return double product.

MSTAR:  goto    DOLST
		DW	DDUP,XORR,ZLESS,TOR
		DW	ABSS,SWAP,ABSS,UMSTA
		DW	RFROM
		DW	QBRAN,MSTAR+13
		DW	DNEGA
MSTA1:		DW	EXIT

;   */MOD	( n1 n2 n3 -- r q )
;		Multiply n1 and n2, then divide by n3. Return mod and quotient.

SSMOD:  goto    DOLST
		DW	TOR,MSTAR,RFROM,MSMOD,EXIT

;   */		( n1 n2 n3 -- q )
;		Multiply n1 by n2, then divide by n3. Return quotient only.

STASL:  goto    DOLST
		DW	SSMOD,SWAP,DROP,EXIT


;; Miscellaneous

;      1+	( a -- a )
;		Add cell size in byte to address.

ONEP:  goto    DOLST
		DW	DOLIT,1,PLUS,EXIT

;      1-	( a -- a )
;		Subtract cell size in byte from address.

ONEM:  goto    DOLST
		DW	DOLIT,-1,PLUS,EXIT

;   CELLS	( n -- n*2 )
;		Multiply tos by cell size in bytes.

CELLS:  goto    DOLST
		DW	DUPP,PLUS,EXIT

;   ALIGNED	( b -- a )
;		Align byte address to the cell boundary.

ALGND:  goto    DOLST
		DW	DOLIT,0,DOLIT,CELLL	;
		DW	UMMOD			;extraByte cellAddress --
ALGN1:		DW	PLUS,EXIT

;   BL		( -- 32 )
;		Return 32, the blank character.

BLANK:  goto    DOLST
		DW	DOLIT,32,EXIT

;   >CHAR	( c -- c )
;		Filter non-printing characters.

TCHAR:  goto    DOLST
		DW	DOLIT,127,ANDD,DUPP	;mask msb
		DW	DOLIT,127,BLANK,WITHI	;check for printable
		DW	QBRAN,TCHAR+14
		DW	DROP,DOLIT,95		;replace non-printables
TCHA1:		DW	EXIT

;   DEPTH	( -- n )
;		Return the depth of the data stack.

DEPTH:  goto    DOLST
		DW	SPAT,SZERO,AT,SUBB
		DW	DOLIT,CELLL,SLASH,EXIT

;   PICK	( ... +n -- ... w )
;		Copy the nth stack item to tos.

PICK:		incf	TL,1		    ; n is zero based index
		rlncf	TL,1		    ; n*2
		bcf	TL,0		    ; clear LSB
		movpf	SP,XL		    ; save SP
		movpf	TL,W
		subwf	SP,1		    ; move SP down
		movpf	SPP,TL		    ; get stack item
		movpf	SPP,TH
		movfp	XL,SP		    ; restore SP
		NEXT


;; Memory access

;   +!		( n a -- )
;		Add n to the contents at address a.

PSTOR:  goto    DOLST
		DW	SWAP,OVER,AT,PLUS
		DW	SWAP,STORE,EXIT

;   2!		( d a -- )
;		Store the double integer to address a.

DSTOR:  goto    DOLST
		DW	SWAP,OVER,STORE
		DW	ONEP,STORE,EXIT

;   2@		( a -- d )
;		Fetch double integer from address a.

DAT:    goto    DOLST
		DW	DUPP,ONEP,AT
		DW	SWAP,AT,EXIT

;   COUNT	( a -- b +n )
;		Given a cell address
;		return count byte of a string and add 1 to byte address.

COUNT:	goto	DOLST
		DW	DUPP,PLUS		; convert to byte address
		DW	DUPP,ONEP
		DW	SWAP,CAT,EXIT

;   HERE	( -- a )
;		Return the top of the code dictionary.

HERE:   goto    DOLST
		DW	CP,AT,EXIT

;   PAD		( -- a )
;		Return the address of a temporary buffer.

PAD:    goto    DOLST
		DW	HERE,DOLIT,80,PLUS,EXIT

;   TIB		( -- a )
;		Return the address of the terminal input buffer.

TIB:    goto    DOLST
		DW	NTIB,ONEP,AT,EXIT

;   @EXECUTE	( a -- )
;		Execute vector stored in address a.

ATEXE:  goto    DOLST
		DW	AT,QDUP 		;?address or zero
		DW	QBRAN,ATEXE+6
		DW	EXECU			;execute if non-zero
EXE1:		DW	EXIT			;do nothing if zero

;   CMOVE	( b1 b2 u -- )
;		Copy u bytes from b1 to b2.

CMOVE:  goto    DOLST
		DW	TOR
		DW	BRAN,CMOVE+12
CMOV1:		DW	TOR,DUPP,CAT
		DW	RAT,CSTOR
		DW	ONEP
		DW	RFROM,ONEP
CMOV2:		DW	DONXT,CMOV1
		DW	DDROP,EXIT

;   FILL	( b u c -- )
;		Fill u bytes of character c to area beginning at b.

FILL:   goto    DOLST
		DW	SWAP,TOR,SWAP
		DW	BRAN,FILL+9
FILL1:		DW	DDUP,CSTOR,ONEP
FILL2:		DW	DONXT,FILL1
		DW	DDROP,EXIT

;   -TRAILING	( b u -- b u )
;		Adjust the count to eliminate trailing white space.

DTRAI:  goto    DOLST
		DW	TOR
		DW	BRAN,DTRAI+15
DTRA1:		DW	BLANK,OVER,RAT,PLUS,CAT,LESS
		DW	QBRAN,DTRAI+15
		DW	RFROM,ONEP,EXIT	;adjusted count
DTRA2:		DW	DONXT,DTRA1
		DW	DOLIT,0,EXIT		;count=0

;   PACK$	( b u a -- a )
;		Build a counted string with u characters from b. Null fill.

PACKS:  goto    DOLST
		DW	DUPP,TOR		;strings only on cell boundary
		DW	DUPP,PLUS		;convert to byte address
		DW	OVER,DOLIT,1,ANDD	;odd length?
		DW	ONEM,QBRAN,PACKS+19	;if odd length, skip
		DW	DDUP,PLUS
		DW	ONEP			;one byte beyond string
		DW	DOLIT,0,SWAP,CSTOR	;append a null byte
PACK2:		DW	DDUP,CSTOR,ONEP 	;save count
		DW	SWAP,CMOVE,RFROM,EXIT	;move string


;; Numeric output, single precision

;   DIGIT	( u -- c )
;		Convert digit u to a character.

DIGIT:  goto    DOLST
		DW	DOLIT,9,OVER,LESS
		DW	DOLIT,7,ANDD,PLUS
		DW	DOLIT,48,PLUS,EXIT

;   EXTRACT	( n base -- n c )
;		Extract the least significant digit from n.

EXTRC:  goto    DOLST
		DW	DOLIT,0,SWAP,UMMOD
		DW	SWAP,DIGIT,EXIT

;   <#		( -- )
;		Initiate the numeric output process.

BDIGS:  goto    DOLST
		DW	PAD,CELLS,HLD,STORE,EXIT

;   HOLD	( c -- )
;		Insert a character into the numeric output string.

HOLD:   goto    DOLST
		DW	HLD,AT,DOLIT,-1,PLUS
		DW	DUPP,HLD,STORE,CSTOR,EXIT

;   #		( u -- u )
;		Extract one digit from u and append the digit to output string.

DIG:    goto    DOLST
		DW	BASE,AT,EXTRC,HOLD,EXIT

;   #S		( u -- 0 )
;		Convert u until all digits are added to the output string.

DIGS:   goto    DOLST
		DW	DIG,DUPP
		DW	QBRAN,DIGS+7
		DW	BRAN,DIGS+1
DIGS2:		DW	EXIT

;   SIGN	( n -- )
;		Add a minus sign to the numeric output string.

SIGN:   goto    DOLST
		DW	ZLESS
		DW	QBRAN,SIGN+7
		DW	DOLIT,45,HOLD
SIGN1:		DW	EXIT

;   #>		( w -- b u )
;		Prepare the output string to be TYPE'd.

EDIGS:  goto    DOLST
		DW	DROP,HLD,AT
		DW	PAD,CELLS,OVER,SUBB,EXIT

;   str		( n -- b u )
;		Convert a signed integer to a numeric string.

STR:    goto    DOLST
		DW	DUPP,TOR,ABSS
		DW	BDIGS,DIGS,RFROM
		DW	SIGN,EDIGS,EXIT

;   HEX		( -- )
;		Use radix 16 as base for numeric conversions.

HEX:    goto    DOLST
		DW	DOLIT,16,BASE,STORE,EXIT

;   DECIMAL	( -- )
;		Use radix 10 as base for numeric conversions.

DECIM:  goto    DOLST
		DW	DOLIT,10,BASE,STORE,EXIT


;; Numeric input, single precision

;   DIGIT?	( c base -- u t )
;		Convert a character to its numeric value. A flag indicates success.

DIGTQ:  goto    DOLST
		DW	TOR,DOLIT,48,SUBB
		DW	DOLIT,9,OVER,LESS
		DW	QBRAN,DIGTQ+19
		DW	DOLIT,7,SUBB
		DW	DUPP,DOLIT,10,LESS,ORR
DGTQ1:		DW	DUPP,RFROM,ULESS,EXIT

;   NUMBER?	( a -- n T | a F )
;		Convert a number string to integer. Push a flag on tos.

NUMBQ:  goto    DOLST
		DW	BASE,AT,TOR,DOLIT,0,OVER,COUNT	;a 0 b n --
		DW	OVER,CAT,DOLIT,36,EQUAL 	;a 0 b n f --
		DW	QBRAN,NUMBQ+20			;if $, hex based
		DW	HEX,SWAP,ONEP
		DW	SWAP,ONEM			;a 0 b+1 n-1 --
NUMQ1:		DW	OVER,CAT,DOLIT,45,EQUAL,TOR	;a 0 b n --
		DW	SWAP,RAT,SUBB,SWAP,RAT,PLUS,QDUP;n=0?
		DW	QBRAN,NUMBQ+70
		DW	ONEM,TOR			;a 0 b --, r:f n-1 --
NUMQ2:		DW	DUPP,TOR,CAT,BASE,AT,DIGTQ	;a 0 # f --
		DW	QBRAN,NUMBQ+63
		DW	SWAP,BASE,AT,STAR,PLUS,RFROM	;a # b --
		DW	ONEP
		DW	DONXT,NUMQ2			;loop back
		DW	RAT,SWAP,DROP			;a # f --, success
		DW	QBRAN,NUMBQ+60
		DW	NEGAT
NUMQ3:		DW	SWAP				;# a --
		DW	BRAN,NUMBQ+69
NUMQ4:		DW	RFROM,RFROM,DDROP,DDROP,DOLIT,0 ;a 0 --, failed
NUMQ5:		DW	DUPP
NUMQ6:		DW	RFROM,DDROP			;# a --
		DW	RFROM,BASE,STORE,EXIT


;; Basic I/O

;   ?KEY	( -- c T | F )
;		Return input character and true, or a false if no input.

QKEY:   goto    DOLST
		DW	TQKEY,ATEXE,EXIT

;   KEY		( -- c )
;		Wait for and return an input character.

KEY:	goto	DOLST
KEY1:		DW	QKEY
		DW	QBRAN,KEY1
		DW	EXIT

;   EMIT	( c -- )
;		Send a character to the output device.

EMIT:   goto    DOLST
		DW	TEMIT,ATEXE,EXIT

;   NUF?	( -- t )
;		Return false if no input, else pause and if CR return true.

NUFQ:   goto    DOLST
		DW	QKEY,DUPP
		DW	QBRAN,NUFQ+10
		DW	DDROP,KEY,DOLIT,CRR,EQUAL
NUFQ1:		DW	EXIT

;   PACE	( -- )
;		Send a pace character for the file downloading process.

PACE:   goto    DOLST
		DW	DOLIT,11,EMIT,EXIT

;   SPACE	( -- )
;		Send the blank character to the output device.

SPACE:  goto    DOLST
		DW	BLANK,EMIT,EXIT

;   SPACES	( +n -- )
;		Send n spaces to the output device.

SPACS:  goto    DOLST
		DW	DOLIT,0,MAX,TOR
		DW	BRAN,SPACS+8
CHAR1:		DW	SPACE
CHAR2:		DW	DONXT,CHAR1
		DW	EXIT

;   TYPE	( b u -- )
;		Output u characters from b.

TYPEE:  goto    DOLST
		DW	TOR
		DW	BRAN,TYPEE+8
TYPE1:		DW	DUPP,CAT,EMIT
		DW	ONEP
TYPE2:		DW	DONXT,TYPE1
		DW	DROP,EXIT

;   CR		( -- )
;		Output a carriage return and a line feed.

CR:     goto    DOLST
		DW	DOLIT,CRR,EMIT
		DW	DOLIT,LF,EMIT,EXIT

;   do$		( -- a )
;		Return the address of a compiled string.

DOSTR:  goto    DOLST
		DW	RFROM,RAT,RFROM,COUNT,PLUS
		DW	ALGND,TOR,SWAP,TOR,EXIT

;   $"|		( -- a )
;		Run time routine compiled by $". Return address of a compiled string.

STRQP:  goto    DOLST
		DW	DOSTR,EXIT		;force a call to do$

;   ."|		( -- )
;		Run time routine of ." . Output a compiled string.

DOTQP:  goto    DOLST
		DW	DOSTR,COUNT,TYPEE,EXIT

;   .R		( n +n -- )
;		Display an integer in a field of n columns, right justified.

DOTR:   goto    DOLST
		DW	TOR,STR,RFROM,OVER,SUBB
		DW	SPACS,TYPEE,EXIT

;   U.R		( u +n -- )
;		Display an unsigned integer in n column, right justified.

UDOTR:  goto    DOLST
		DW	TOR,BDIGS,DIGS,EDIGS
		DW	RFROM,OVER,SUBB
		DW	SPACS,TYPEE,EXIT

;   U.		( u -- )
;		Display an unsigned integer in free format.

UDOT:   goto    DOLST
		DW	BDIGS,DIGS,EDIGS
		DW	SPACE,TYPEE,EXIT

;   .		( w -- )
;		Display an integer in free format, preceeded by a space.

DOT:    goto    DOLST
		DW	BASE,AT,DOLIT,10,XORR	;?decimal
		DW	QBRAN,DOT+10
		DW	UDOT,EXIT		;no, display unsigned
DOT1:		DW	STR,SPACE,TYPEE,EXIT	;yes, display signed

;   ?		( a -- )
;		Display the contents in a memory cell.

QUEST:  goto    DOLST
		DW	AT,DOT,EXIT


;; Parsing

;   parse	( b u c -- b u delta ; <string> )
;		Scan string delimited by c. Return found string and its offset.

PARS:   goto    DOLST
		DW	TEMP,STORE,OVER,TOR,DUPP
		DW	QBRAN,PARS+69
		DW	ONEM,TEMP,AT,BLANK,EQUAL
		DW	QBRAN,PARS+34
		DW	TOR
PARS1:		DW	BLANK,OVER,CAT		;skip leading blanks ONLY
		DW	SUBB,ZLESS,INVER
		DW	QBRAN,PARS+33
		DW	ONEP
		DW	DONXT,PARS1
		DW	RFROM,DROP,DOLIT,0,DUPP,EXIT
PARS2:		DW	RFROM
PARS3:		DW	OVER,SWAP
		DW	TOR
PARS4:		DW	TEMP,AT,OVER,CAT,SUBB	;scan for delimiter
		DW	TEMP,AT,BLANK,EQUAL
		DW	QBRAN,PARS+49
		DW	ZLESS
PARS5:		DW	QBRAN,PARS+58
		DW	ONEP
		DW	DONXT,PARS4
		DW	DUPP,TOR
		DW	BRAN,PARS+63
PARS6:		DW	RFROM,DROP,DUPP
		DW	ONEP,TOR
PARS7:		DW	OVER,SUBB
		DW	RFROM,RFROM,SUBB,EXIT
PARS8:		DW	OVER,RFROM,SUBB,EXIT

;   PARSE	( c -- b u ; <string> )
;		Scan input stream and return counted string delimited by c.

PARSE:  goto    DOLST
		DW	TOR,TIB,INN,AT,PLUS	;current input buffer pointer
		DW	NTIB,AT,INN,AT,SUBB	;remaining count
		DW	RFROM,PARS,INN,PSTOR,EXIT

;   .(		( -- )
;		Output following string up to next ) .

DOTPR:  goto    DOLST
		DW	DOLIT,41,PARSE,TYPEE,EXIT

;   (		( -- )
;		Ignore following string up to next ) . A comment.

PAREN:  goto    DOLST
		DW	DOLIT,41,PARSE,DDROP,EXIT

;   \		( -- )
;		Ignore following text till the end of line.

BKSLA:  goto    DOLST
		DW	NTIB,AT,INN,STORE,EXIT

;   CHAR	( -- c )
;		Parse next word and return its first character.

CHAR:   goto    DOLST
		DW	BLANK,PARSE,DROP,CAT,EXIT

;   TOKEN	( -- a ; <string> )
;		Parse a word from input stream and copy it to name dictionary.

TOKEN:  goto    DOLST
		DW	BLANK,PARSE,DOLIT,31,MIN
		DW	NP,AT,OVER
		DW	DOLIT,CELLL,SLASH
		DW	SUBB,ONEM
		DW	PACKS,EXIT

;   WORD	( c -- a ; <string> )
;		Parse a word from input stream and copy it to code dictionary.

WORDD:	 goto	 DOLST
		DW	PARSE,HERE,PACKS,EXIT


;; Dictionary search

;   NAME>	( na -- ca )
;		Return a code address given a name address.

NAMET:  goto    DOLST
		DW	ONEM,ONEM,AT,EXIT

;   SAME?	( a a u -- a a f \ -0+ )
;		Compare u cells in two strings. Return 0 if identical.

SAMEQ:  goto    DOLST
		DW	TOR
		DW	BRAN,SAMEQ+19
SAME1:		DW	OVER,RAT,PLUS,AT
		DW	OVER,RAT,PLUS,AT
		DW	XORR,QDUP
		DW	QBRAN,SAMEQ+19
		DW	RFROM,DROP,EXIT		;strings not equal
SAME2:		DW	DONXT,SAME1
		DW	DOLIT,0,EXIT		;strings equal

;   find	( a va -- ca na | a F )
;		Search a vocabulary for a string. Return ca and na if succeeded.

FIND:   goto    DOLST
		DW	SWAP,DUPP,DUPP,PLUS,CAT     ;va a n --
		DW	DOLIT,CELLL,SLASH,TEMP,STORE;va a --
		DW	DUPP,AT,TOR,ONEP,SWAP	    ;a+1 va --
FIND1:		DW	AT,DUPP 		    ;a+1 na na --
		DW	QBRAN,FIND+40
		DW	DUPP,AT,DOLIT,MASKK,ANDD,RAT,XORR; comp 1st cell
		DW	QBRAN,FIND+34		    ;a+1 na --
		DW	ONEP,DOLIT,-1		    ;not same, a+1 na+1 -1 --
		DW	BRAN,FIND+38
FIND2:		DW	ONEP,TEMP,AT,SAMEQ	    ;comp rest of name
FIND3:		DW	BRAN,FIND+46		    ;a+1 na+1 f --
FIND6:		DW	RFROM,DROP		    ;na=0, end of voc
		DW	SWAP,ONEM,SWAP,EXIT	    ;a 0 --, failed
FIND4:		DW	QBRAN,FIND+52		    ;sucess, goto FIND5
		DW	ONEM,ONEM		    ;not same, a+1 la --
		DW	BRAN,FIND1		    ;loop, comp next name
FIND5:		DW	RFROM,DROP,SWAP,DROP	    ;found, na+1 --
		DW	ONEM			    ;na --
		DW	DUPP,NAMET,SWAP,EXIT	    ;ca na --

;   NAME?	( a -- ca na | a F )
;		Search all context vocabularies for a string.

NAMEQ:  goto    DOLST
		DW	CNTXT,DUPP,DAT,XORR	;?context=also
		DW	QBRAN,NAMEQ+8
		DW	ONEM			;no, start with context
NAMQ1:		DW	TOR
NAMQ2:		DW	RFROM,ONEP,DUPP,TOR	;next in search order
		DW	AT,QDUP
		DW	QBRAN,NAMEQ+24
		DW	FIND,QDUP		;search vocabulary
		DW	QBRAN,NAMQ2
		DW	RFROM,DROP,EXIT		;found name
NAMQ3:		DW	RFROM,DROP		;name not found
		DW	DOLIT,0,EXIT		;false flag


;; Terminal response

;   ^H		( bot eot cur -- bot eot cur )
;		Backup the cursor by one character.

BKSP:   goto    DOLST
		DW	TOR,OVER,RFROM,SWAP,OVER,XORR
		DW	QBRAN,BKSP+21
		DW	DOLIT,BKSPP,TECHO,ATEXE,ONEM
		DW	BLANK,TECHO,ATEXE
		DW	DOLIT,BKSPP,TECHO,ATEXE
BACK1:		DW	EXIT

;   TAP		( bot eot cur c -- bot eot cur )
;		Accept and echo the key stroke and bump the cursor.

TAP:    goto    DOLST
		DW	DUPP,TECHO,ATEXE
		DW	OVER,CSTOR,ONEP,EXIT

;   kTAP	( bot eot cur c -- bot eot cur )
;		Process a key stroke, CR or backspace.

KTAP:	goto	DOLST
		DW	DUPP,DOLIT,CRR,XORR
		DW	QBRAN,KTAP+17
		DW	DOLIT,BKSPP,XORR
		DW	QBRAN,KTAP+15
		DW	BLANK,TAP,EXIT
KTAP1:		DW	BKSP,EXIT
KTAP2:		DW	DROP,SWAP,DROP,DUPP,EXIT

;   accept	( b u -- b u )
;		Accept characters to input buffer. Return with actual count.

ACCEP:  goto    DOLST
		DW	OVER,PLUS,OVER		    ;b b+u b --
ACCP1:		DW	DDUP,XORR		    ;b b+u b f --
		DW	QBRAN,ACCEP+23		    ;buffer full, exit
		DW	KEY,DUPP		    ;b b+u b c c --
		DW	BLANK,DOLIT,127,WITHI	    ;b b+u b c f --
		DW	QBRAN,ACCEP+19
		DW	TAP			    ;b b+u b+1 --
		DW	BRAN,ACCEP+21
ACCP2:		DW	TTAP,ATEXE
ACCP3:		DW	BRAN,ACCP1		    ;loop
ACCP4:		DW	DROP,OVER,SUBB,EXIT	    ; b u --

;   EXPECT	( b u -- )
;		Accept input stream and store count in SPAN.

EXPEC:  goto    DOLST
		DW	TEXPE,ATEXE,SPAN,STORE,DROP,EXIT

;   QUERY	( -- )
;		Accept input stream to terminal input buffer.

QUERY:  goto    DOLST
		DW	TIB
		DW	DOLIT,80,TEXPE,ATEXE,NTIB,STORE
		DW	DROP,DOLIT,0,INN,STORE,EXIT


;; Error handling

;   CATCH	( ca -- 0 | err# )
;		Execute word at ca and set up an error frame for it.

CATCH:  goto    DOLST
		DW	SPAT,TOR,HANDL,AT,TOR	;save error frame
		DW	RPAT,HANDL,STORE,EXECU	;execute
		DW	RFROM,HANDL,STORE	;restore error frame
		DW	RFROM,DROP,DOLIT,0,EXIT	;no error

;   THROW	( err# -- err# )
;		Reset system to current local error frame an update error flag.

THROW:  goto    DOLST
		DW	HANDL,AT,RPSTO		;restore return stack
		DW	RFROM,HANDL,STORE	;restore handler frame
		DW	RFROM,SWAP,TOR,SPSTO	;restore data stack
		DW	DROP,RFROM,EXIT

;   NULL$	( -- a )
;		Return address of a null string with zero count.

NULLS:	goto	DOVAR
		DW	0,0

;   ABORT	( -- )
;		Reset data stack and jump to QUIT.

ABORT:  goto    DOLST
		DW	NULLS,THROW

;   abort"	( f -- )
;		Run time routine of ABORT" . Abort with a message.

ABORQ:  goto    DOLST
		DW	QBRAN,ABORQ+5		;text flag
		DW	DOSTR,THROW		;pass error string
ABOR1:		DW	DOSTR,DROP,EXIT		;drop error


;; The text interpreter

;   $INTERPRET	( a -- )
;		Interpret a word. If failed, try to convert it to an integer.

INTER:  goto    DOLST
		DW	NAMEQ,QDUP		;?defined
		DW	QBRAN,INTER+19
		DW	AT,DOLIT,COMPO,ANDD	;?compile only lexicon bits
		DW	ABORQ,0xD20,"compile only"
		DW	EXECU,EXIT		;execute defined word
INTE1:		DW	TNUMB,ATEXE		;convert a number
		DW	QBRAN,INTER+24
		DW	EXIT
INTE2:		DW	THROW			;error

;   [		( -- )
;		Start the text interpreter.

LBRAC:  goto    DOLST
		DW	DOLIT,INTER,TEVAL,STORE,EXIT

;   .OK		( -- )
;		Display "ok" only while interpreting.

DOTOK:  goto    DOLST
		DW	DOLIT,INTER,TEVAL,AT,EQUAL
		DW	QBRAN,DOTOK+11
		DW	DOTQP,0x320,"ok"       ; ok
DOTO1:		DW	CR,EXIT

;   ?STACK	( -- )
;		Abort if the data stack underflows.

QSTAC:  goto    DOLST
		DW	DEPTH,ZLESS		;check only for underflow
		DW	ABORQ,0XA20,"underflow" ;underflow
		DW	EXIT

;   EVAL	( -- )
;		Interpret the input stream.

EVAL:   goto    DOLST
EVAL1:		DW	TOKEN,DUPP
		DW	DUPP,PLUS,CAT		;?input stream empty
		DW	QBRAN,EVAL+13
		DW	TEVAL,ATEXE,QSTAC	;evaluate input, check stack
		DW	BRAN,EVAL1
EVAL2:		DW	DROP,TPROM,ATEXE,EXIT	;prompt

;   Shell

;   PRESET	( -- )
;		Reset data stack pointer and the terminal input buffer.

PRESE:  goto    DOLST
		DW	SZERO,AT,SPSTO
		DW	DOLIT,TIBB,NTIB,ONEP,STORE,EXIT

;   xio		( a a a -- )
;		Reset the I/O vectors 'EXPECT, 'TAP, 'ECHO and 'PROMPT.

XIO:    goto    DOLST
		DW	DOLIT,ACCEP,TEXPE,DSTOR
		DW	TECHO,DSTOR,EXIT

;   FILE	( -- )
;		Select I/O vectors for file download.

FILE:   goto    DOLST
		DW	DOLIT,PACE,DOLIT,DROP
		DW	DOLIT,KTAP,XIO,EXIT

;   HAND	( -- )
;		Select I/O vectors for terminal interface.

HAND:   goto    DOLST
		DW	DOLIT,DOTOK,DOLIT,EMIT
		DW	DOLIT,KTAP,XIO,EXIT

;   I/O		( -- a )
;		Array to store default I/O vectors.

ISLO:	goto	DOVAR
		DW	QRX,TXSTO		;default I/O vectors

;   CONSOLE	( -- )
;		Initiate terminal interface.

CONSO:  goto    DOLST
		DW	ISLO,DAT,TQKEY,DSTOR	;restore default I/O device
		DW	HAND,EXIT		;keyboard input

;   QUIT	( -- )
;		Reset return stack pointer and start text interpreter.

QUIT:   goto    DOLST
		DW	RZERO,AT,RPSTO		;reset return stack pointer
QUIT1:		DW	LBRAC			;start interpretation
QUIT2:		DW	QUERY			;get input
		DW	DOLIT,EVAL,CATCH,QDUP	;evaluate input
		DW	QBRAN,QUIT2		;continue till error
		DW	TPROM,AT,SWAP		;save input device
		DW	CONSO,NULLS,OVER,XORR	;?display error message
		DW	QBRAN,QUIT+26
		DW	SPACE,COUNT,TYPEE	;error message
		DW	DOTQP,0x13F		;error prompt ?
QUIT3:		DW	DOLIT,DOTOK,XORR	;?file input
		DW	QBRAN,QUIT+34
		DW	DOLIT,ERR,EMIT		;file error, tell host
QUIT4:		DW	PRESE			;some cleanup
		DW	BRAN,QUIT1


;; The compiler

;   '           ( -- ca )
;		Search context vocabularies for the next word in input stream.

TICK:   goto    DOLST
		DW	TOKEN,NAMEQ		;?defined
		DW	QBRAN,TICK+6
		DW	EXIT			;yes, push code address
TICK1:		DW	THROW			;no, error

;   ALLOT	( n -- )
;		Allocate n cells to the code dictionary.

ALLOT:  goto    DOLST
		DW	CP,PSTOR,EXIT		;adjust code pointer

;   ,		( w -- )
;		Compile an integer into the code dictionary.

COMMA:  goto    DOLST
		DW	HERE,DUPP,ONEP 	;cell boundary
		DW	CP,STORE,STORE,EXIT	;adjust code pointer, compile

;   [COMPILE]	( -- ; <string> )
;		Compile the next immediate word into code dictionary.

BCOMP:  goto    DOLST
		DW	TICK,COMMA,EXIT

;   COMPILE	( -- )
;		Compile the next address in colon list to code dictionary.

COMPI:  goto    DOLST
		DW	RFROM,DUPP,AT,COMMA	;compile address
		DW	ONEP,TOR,EXIT		;adjust return address

;   LITERAL	( w -- )
;		Compile tos to code dictionary as an integer literal.

LITER:  goto    DOLST
		DW	COMPI,DOLIT,COMMA,EXIT

;   $,"		( -- )
;		Compile a literal string up to next " .

STRCQ:  goto    DOLST
		DW	DOLIT,34,WORDD		;move string to code dictionary
		DW	COUNT,PLUS,ALGND	;calculate aligned end of string
		DW	CP,STORE,EXIT		;adjust the code pointer

;   RECURSE	( -- )
;		Make the current word available for compilation.

RECUR:  goto    DOLST
		DW	LAST,AT,NAMET,COMMA,EXIT


;; Structures

;   FOR		( -- a )
;		Start a FOR-NEXT loop structure in a colon definition.

FOR:    goto    DOLST
		DW	COMPI,TOR,HERE,EXIT

;   BEGIN	( -- a )
;		Start an infinite or indefinite loop structure.

BEGIN:  goto    DOLST
		DW	HERE,EXIT

;   NEXT	( a -- )
;		Terminate a FOR-NEXT loop structure.

NEXTT:  goto    DOLST
		DW	COMPI,DONXT,COMMA,EXIT

;   UNTIL	( a -- )
;		Terminate a BEGIN-UNTIL indefinite loop structure.

UNTIL:  goto    DOLST
		DW	COMPI,QBRAN,COMMA,EXIT

;   AGAIN	( a -- )
;		Terminate a BEGIN-AGAIN infinite loop structure.

AGAIN:  goto    DOLST
		DW	COMPI,BRAN,COMMA,EXIT

;   IF		( -- A )
;		Begin a conditional branch structure.

IFF:    goto    DOLST
		DW	COMPI,QBRAN,HERE
		DW	DOLIT,0,COMMA,EXIT

;   AHEAD	( -- A )
;		Compile a forward branch instruction.

AHEAD:  goto    DOLST
		DW	COMPI,BRAN,HERE,DOLIT,0,COMMA,EXIT

;   REPEAT	( A a -- )
;		Terminate a BEGIN-WHILE-REPEAT indefinite loop.

REPEA:  goto    DOLST
		DW	AGAIN,HERE,SWAP,STORE,EXIT

;   THEN	( A -- )
;		Terminate a conditional branch structure.

THENN:  goto    DOLST
		DW	HERE,SWAP,STORE,EXIT

;   AFT		( a -- a A )
;		Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.

AFT:    goto    DOLST
		DW	DROP,AHEAD,BEGIN,SWAP,EXIT

;   ELSE	( A -- A )
;		Start the false clause in an IF-ELSE-THEN structure.

ELSEE:  goto    DOLST
		DW	AHEAD,SWAP,THENN,EXIT

;   WHILE	( a -- A a )
;		Conditional branch out of a BEGIN-WHILE-REPEAT loop.

WHILE:  goto    DOLST
		DW	IFF,SWAP,EXIT

;   ABORT"	( -- ; <string> )
;		Conditional abort with an error message.

ABRTQ:  goto    DOLST
		DW	COMPI,ABORQ,STRCQ,EXIT

;   $"		( -- ; <string> )
;		Compile an inline string literal.

STRQ:   goto    DOLST
		DW	COMPI,STRQP,STRCQ,EXIT

;   ."		( -- ; <string> )
;		Compile an inline string literal to be typed out at run time.

DOTQ:   goto    DOLST
		DW	COMPI,DOTQP,STRCQ,EXIT


;; Name compiler

;   ?UNIQUE	( a -- a )
;		Display a warning message if the word already exists.

UNIQU:  goto    DOLST
		DW	DUPP,NAMEQ		;?name exists
		DW	QBRAN,UNIQU+13		;redefinitions are OK
		DW	DOTQP,0x720,"reDef "   ;
		DW	OVER,COUNT,TYPEE	;just in case its not planned
UNIQ1:		DW	DROP,EXIT

;   $,n		( na -- )
;		Build a new dictionary name using the string at na.

SNAME:  goto    DOLST
		DW	DUPP,DUPP,PLUS,CAT	;?null input
		DW	QBRAN,SNAME+25
		DW	UNIQU			;?redefinition
		DW	DUPP,LAST,STORE		;save na for vocabulary link
		DW	HERE,SWAP		;here na --
		DW	ONEM			;here la --
		DW	CRRNT,AT,AT,OVER,STORE	;fix vocabulary pointer
		DW	ONEM			;here ca --
		DW	DUPP,NP,STORE		;adjust name pointer
		DW	STORE,EXIT		;save code pointer
PNAM1:		DW	STRQP,0x520,"name"	;null input
		DW	THROW


;; FORTH compiler

;   $COMPILE	( a -- )
;		Compile next word to code dictionary as a token or literal.

SCOMP:  goto    DOLST
		DW	NAMEQ,QDUP		;?defined
		DW	QBRAN,SCOMP+15
		DW	AT,DOLIT,IMEDD,ANDD	;?immediate
		DW	QBRAN,SCOMP+13
		DW	EXECU,EXIT		;its immediate, execute
SCOM1:		DW	COMMA,EXIT		;its not immediate, compile
SCOM2:		DW	TNUMB,ATEXE		;try to convert to number
		DW	QBRAN,SCOMP+21
		DW	LITER,EXIT		;compile number as integer
SCOM3:		DW	THROW			;error

;   OVERT	( -- )
;		Link a new word into the current vocabulary.

OVERT:  goto    DOLST
		DW	LAST,AT,CRRNT,AT,STORE,EXIT

;   ;		( -- )
;		Terminate a colon definition.

SEMIS:  goto    DOLST
		DW	COMPI,EXIT,LBRAC,OVERT,EXIT

;   ]		( -- )
;		Start compiling the words in the input stream.

RBRAC:  goto    DOLST
		DW	DOLIT,SCOMP,TEVAL,STORE,EXIT

;   call,	( -- )
;		Assemble a 'goto DOLST' instruction for a colon word.

CALLC:  goto    DOLST
		DW	DOLIT,CALLL,COMMA	;Direct Threaded Code
		DW	EXIT

;   :   		( -- ; <string> )
;		Start a new colon definition using next word as its name.

COLON:  goto    DOLST
		DW	TOKEN,SNAME
		DW	CALLC,RBRAC,EXIT

;   IMMEDIATE	( -- )
;		Make the last compiled word an immediate word.

IMMED:  goto    DOLST
		DW	DOLIT,IMEDD,LAST,AT,AT,ORR
		DW	LAST,AT,STORE,EXIT


;; Defining words

;   CONSTANT	( u -- ; <string> )
;		Compile a new user variable.

CONST:	goto	DOLST
		DW	TOKEN,SNAME,OVERT
		DW	COMPI,CONNN,COMMA,EXIT

;   USER	( u -- ; <string> )
;		Compile a new user variable.

USER:   goto    DOLST
		DW	TOKEN,SNAME,OVERT
		DW	COMPI,USEEE,COMMA,EXIT

;   CREATE	( -- ; <string> )
;		Compile a new array entry without allocating code space.

CREAT:  goto    DOLST
		DW	TOKEN,SNAME,OVERT
		DW	COMPI,VARRR,EXIT

;   VARIABLE	( -- ; <string> )
;		Compile a new variable initialized to 0.

VARIA:  goto    DOLST
		DW	CREAT,DOLIT,0,COMMA,EXIT


;;   Tools

;   _TYPE	( b u -- )
;		Display a string. Filter non-printing characters.

UTYPE:	goto	DOLST
		DW	TOR			;start count down loop
		DW	BRAN,UTYPE+9		;skip first pass
UTYP1:		DW	DUPP,CAT,TCHAR,EMIT	;display only printable
		DW	ONEP			;increment address
UTYP2:		DW	DONXT,UTYP1		;loop till done
		DW	DROP,EXIT

;   dmp 	( b u -- )
;		Dump u bytes from b.

DMP:    goto    DOLST
		DW	TOR			;start count down loop
		DW	BRAN,DMP+10		;skip first pass
PDUM1:		DW	DUPP,CAT,DOLIT,3,UDOTR	;display numeric data
		DW	ONEP			;increment address
PDUM2:		DW	DONXT,PDUM1		;loop till done
		DW	DROP,EXIT

;   DUMP	( a u -- )
;		Dump u bytes from a, in a formatted manner.

DUMP:   goto    DOLST
		DW	BASE,AT,TOR,HEX 	;save radix, set hex
		DW	DOLIT,16,SLASH		;change count to lines
		DW	TOR			;start count down loop
DUMP1:		DW	CR,DUPP,DOLIT,4,UDOTR	;print header
		DW	DUPP,DUPP,PLUS,SPACE	;get byte address
		DW	DOLIT,16,DDUP,DMP	;display numeric
		DW	SPACE,SPACE,UTYPE	;display printable characters
		DW	NUFQ,INVER		;user control
		DW	QBRAN,DUMP+36
		DW	DOLIT,8,PLUS		;bump address
		DW	DONXT,DUMP1		;loop till done
		DW	BRAN,DUMP+38
DUMP2:		DW	RFROM,DROP		;cleanup loop stack, early exit
DUMP3:		DW	DROP,RFROM,BASE,STORE	;restore radix
		DW	EXIT

;   .S		( ... -- ... )
;		Display the contents of the data stack.

DOTS:   goto    DOLST
		DW	CR,DEPTH		;stack depth
		DW	TOR			;start count down loop
		DW	BRAN,DOTS+9		;skip first pass
DOTS1:		DW	RAT,PICK,DOT		;index stack, display contents
DOTS2:		DW	DONXT,DOTS1		;loop till done
		DW	DOTQP,0x420,"<sp"	;<sp
		DW	EXIT

;   !CSP	( -- )
;		Save stack pointer in CSP for error checking.

STCSP:  goto    DOLST
		DW	SPAT,CSP,STORE,EXIT	;save pointer

;   ?CSP	( -- )
;		Abort if stack pointer differs from that saved in CSP.

QCSP:   goto    DOLST
		DW	SPAT,CSP,AT,XORR	;compare pointers
		DW	ABORQ,0x720,"stacks"	;abort if different
		DW	EXIT

;   >NAME	( ca -- na | F )
;		Convert code address to a name address.

TNAME:  goto    DOLST
		DW	CRRNT			;vocabulary link
TNAM1:		DW	ONEP,AT,QDUP		;check all vocabularies
		DW	QBRAN,TNAME+30
		DW	DDUP
TNAM2:		DW	AT,DUPP			;?last word in a vocabulary
		DW	QBRAN,TNAME+20
		DW	DDUP,NAMET,XORR		;compare
		DW	QBRAN,TNAME+20
		DW	ONEM			;continue with next word
		DW	BRAN,TNAM2
TNAM3:		DW	SWAP,DROP,QDUP
		DW	QBRAN,TNAM1
		DW	SWAP,DROP,SWAP,DROP,EXIT
TNAM4:		DW	DROP,DOLIT,0,EXIT	;false flag

;   .ID		( na -- )
;		Display the name at address.

DOTID:  goto    DOLST
		DW	QDUP			;if zero no name
		DW	QBRAN,DOTID+10
		DW	COUNT,DOLIT,31,ANDD	;mask lexicon bits
		DW	UTYPE,EXIT		;display name string
DOTI1:		DW	DOTQP,0x920,"{noName}"
		DW	EXIT

;   SEE		( -- ; <string> )
;		A simple decompiler.

SEE:    goto    DOLST
		DW	TICK			;starting address
		DW	CR
SEE1:		DW	ONEP		;next word
		DW	DUPP,AT,DUPP		;?does it contain a zero
		DW	QBRAN,SEE+10
		DW	TNAME			;?is it a name
SEE2:		DW	QDUP			;name address or zero
		DW	QBRAN,SEE+17
		DW	SPACE,DOTID		;display name
		DW	BRAN,SEE+20
SEE3:		DW	DUPP,AT,UDOT		;display number
SEE4:		DW	NUFQ			;user control
		DW	QBRAN,SEE1
		DW	DROP,EXIT

;   WORDS	( -- )
;		Display the names in the context vocabulary.

WORDS:  goto    DOLST
		DW	CR,CNTXT,AT		;only in context
WORS1:		DW	AT,QDUP			;?at end of list
		DW	QBRAN,WORDS+16
		DW	DUPP,SPACE,DOTID	;display a name
		DW	ONEM		;link field address
		DW	NUFQ			;user control
		DW	QBRAN,WORS1
		DW	DROP
WORS2:		DW	EXIT


;; Hardware reset

;   VER		( -- n )
;		Return the version number of this implementation.

VERSN:  goto    DOLST
		DW	DOLIT,VER*256+EXT,EXIT

;   hi		( -- )
;		Display the sign-on message of eForth.

HI:     goto    DOLST
		DW	STOIO,CR		;initialize I/O
		DW	DOTQP,0x865,"Forth v"	;model
		DW	BASE,AT,HEX		;save radix
		DW	VERSN,BDIGS,DIG,DIG
		DW	DOLIT,46,HOLD
		DW	DIGS,EDIGS,TYPEE	;format version number
		DW	BASE,STORE,CR,EXIT	;restore radix

;   'BOOT       ( -- a )
;		The application startup vector.

TBOOT:	goto	DOVAR
		DW	HI			;application to boot

;   COLD	( -- )
;		The hilevel cold start sequence.

COLD:	goto	DOLST
;		DW	DOLIT,UZERO,DOLIT,UPP
;		DW	DOLIT,37,MOVE		;initialize user area
COLD1:		DW	DOLIT,0X28,DOLIT,0X4028 ;init upper goto area
		DW	DOLIT,0X38,MOVE
		DW	PRESE			;initialize stack and TIB
		DW	TBOOT,ATEXE		;application boot
		DW	FORTH,CNTXT,AT,DUPP	;initialize search order
		DW	CRRNT,DSTOR,OVERT
		DW	QUIT			;start interpretation
		DW	BRAN,COLD		;just in case


;;	name dictionary

;	The name dictionary is compiled separately because of
;	the limitiations in the macro assembler to handle forward
;	references.  It is compiled immediately after the code
;	dictionary in ROM.  The linking is in the forward order
;	inverse from that in the regular eForth.  However, the
;	order is purely incidental and has no effect on the system
;	operation.  It can be reversed by reversing the assembling
;	list.
;
;	In the user defined name dictionary in the RAM memory,
;	word names are linked in the reversed order, growing from
;	high to low memory.

	VARIABLE	LK=	0

_CODE:	MACRO	LEX,STRING,POINTER
	DW	POINTER 		; fill code pointer field
	DW	LINK#V(LK)		; fill link field
	LK+=1				; make a new label
LINK#V(LK)	DW  LEX 		; name length and lexicon
	DW	STRING			; fill name field
        ENDM
LINK0	EQU	0

		EXPAND			; expand name macros

		_CODE	0X33F,"RX",QRX
		_CODE	0X354,"X!",TXSTO
		_CODE	0X321,"IO",STOIO
		_CODE	COMPO+0X564,"oLIT",DOLIT
		_CODE	COMPO+0X664,"oLIST",DOLST
		_CODE	0X445,"XIT",EXIT
		_CODE	0X745,"XECUTE",EXECU
		_CODE	COMPO+0X46E,"ext",DONXT
		_CODE	COMPO+0X73F,"branch",QBRAN
		_CODE	COMPO+0X662,"ranch",BRAN
		_CODE	0X121,"",STORE
		_CODE	0X140,"",AT
		_CODE	0X243,"!",CSTOR
		_CODE	0X243,"@",CAT
		_CODE	0X352,"P@",RPAT
		_CODE	COMPO+0X352,"P!",RPSTO
		_CODE	0X252,">",RFROM
		_CODE	0X252,"@",RAT
		_CODE	COMPO+0X23E,"R",TOR
		_CODE	0X353,"P@",SPAT
		_CODE	0X353,"P!",SPSTO
		_CODE	0X444,"ROP",DROP
		_CODE	0X344,"UP",DUPP
		_CODE	0X453,"WAP",SWAP
		_CODE	0X44F,"VER",OVER
		_CODE	0X230,"<",ZLESS
		_CODE	0X341,"ND",ANDD
		_CODE	0X24F,"R",ORR
		_CODE	0X358,"OR",XORR
		_CODE	0X355,"M+",UPLUS
		_CODE	0X44D,"OVE",MOVE
		_CODE	COMPO+0X564,"oVAR",DOVAR
		_CODE	0X255,"P",UP
		_CODE	COMPO+0X664,"oUSER",DOUSE
		_CODE	0X353,"P0",SZERO
		_CODE	0X352,"P0",RZERO
		_CODE	0X53F,"KEY",TQKEY
		_CODE	0X527,"EMIT",TEMIT
		_CODE	0X727,"EXPECT",TEXPE
		_CODE	0X427,"TAP",TTAP
		_CODE	0X527,"ECHO",TECHO
		_CODE	0X727,"PROMPT",TPROM
		_CODE	0X442,"ASE",BASE
		_CODE	COMPO+0X374,"mp",TEMP
		_CODE	0X453,"PAN",SPAN
		_CODE	0X33E,"IN",INN
		_CODE	0X423,"TIB",NTIB
		_CODE	0X343,"SP",CSP
		_CODE	0X527,"EVAL",TEVAL
		_CODE	0X727,"NUMBER",TNUMB
		_CODE	0X348,"LD",HLD
		_CODE	0X748,"ANDLER",HANDL
		_CODE	0X743,"ONTEXT",CNTXT
		_CODE	0X743,"URRENT",CRRNT
		_CODE	0X243,"P",CP
		_CODE	0X24E,"P",NP
		_CODE	0X44C,"AST",LAST
		_CODE	COMPO+0X564,"oVOC",DOVOC
		_CODE	0X546,"ORTH",FORTH
		_CODE	0X43F,"DUP",QDUP
		_CODE	0X352,"OT",ROT
		_CODE	0X532,"DROP",DDROP
		_CODE	0X432,"DUP",DDUP
		_CODE	0X12B,"",PLUS
		_CODE	0X34E,"OT",INVER
		_CODE	0X64E,"EGATE",NEGAT
		_CODE	0X744,"NEGATE",DNEGA
		_CODE	0X12D,"",SUBB
		_CODE	0X341,"BS",ABSS
		_CODE	0X13D,"",EQUAL
		_CODE	0X255,"<",ULESS
		_CODE	0X13C,"",LESS
		_CODE	0X34D,"AX",MAX
		_CODE	0X34D,"IN",MIN
		_CODE	0X657,"ITHIN",WITHI
		_CODE	0X655,"M/MOD",UMMOD
		_CODE	0X54D,"/MOD",MSMOD
		_CODE	0X42F,"MOD",SLMOD
		_CODE	0X34D,"OD",MODD
		_CODE	0X12F,"",SLASH
		_CODE	0X355,"M*",UMSTA
		_CODE	0X12A,"",STAR
		_CODE	0X24D,"*",MSTAR
		_CODE	0X52A,"/MOD",SSMOD
		_CODE	0X22A,"/",STASL
		_CODE	0X231,"+",ONEP
		_CODE	0X231,"-",ONEM
		_CODE	0X543,"ELLS",CELLS
		_CODE	0X741,"LIGNED",ALGND
		_CODE	0X242,"L",BLANK
		_CODE	0X53E,"CHAR",TCHAR
		_CODE	0X544,"EPTH",DEPTH
		_CODE	0X450,"ICK",PICK
		_CODE	0X22B,"!",PSTOR
		_CODE	0X232,"!",DSTOR
		_CODE	0X232,"@",DAT
		_CODE	0X543,"OUNT",COUNT
		_CODE	0X448,"ERE",HERE
		_CODE	0X350,"AD",PAD
		_CODE	0X354,"IB",TIB
		_CODE	0X840,"EXECUTE",ATEXE
		_CODE	0X543,"MOVE",CMOVE
		_CODE	0X446,"ILL",FILL
		_CODE	0X92D,"TRAILING",DTRAI
		_CODE	0X550,"ACK$",PACKS
		_CODE	0X544,"IGIT",DIGIT
		_CODE	0X745,"XTRACT",EXTRC
		_CODE	0X23C,"#",BDIGS
		_CODE	0X448,"OLD",HOLD
		_CODE	0X123,"",DIG
		_CODE	0X223,"S",DIGS
		_CODE	0X453,"IGN",SIGN
		_CODE	0X223,">",EDIGS
		_CODE	0X373,"tr",STR
		_CODE	0X348,"EX",HEX
		_CODE	0X744,"ECIMAL",DECIM
		_CODE	0X644,"IGIT?",DIGTQ
		_CODE	0X74E,"UMBER?",NUMBQ
		_CODE	0X43F,"KEY",QKEY
		_CODE	0X34B,"EY",KEY
		_CODE	0X445,"MIT",EMIT
		_CODE	0X44E,"UF?",NUFQ
		_CODE	0X450,"ACE",PACE
		_CODE	0X553,"PACE",SPACE
		_CODE	0X653,"PACES",SPACS
		_CODE	0X454,"YPE",TYPEE
		_CODE	0X243,"R",CR
		_CODE	COMPO+0X364,"o$",DOSTR
		_CODE	COMPO+0X324,0X227C,STRQP
		_CODE	COMPO+0X32E,0X227C,DOTQP
		_CODE	0X22E,"R",DOTR
		_CODE	0X355,".R",UDOTR
		_CODE	0X255,".",UDOT
		_CODE	0X12E,"",DOT
		_CODE	0X13F,"",QUEST
		_CODE	0X570,"arse",PARS
		_CODE	0X550,"ARSE",PARSE
		_CODE	IMEDD+0X22E,"(",DOTPR
		_CODE	IMEDD+0X128,"",PAREN
		_CODE	IMEDD+0X15C,"",BKSLA
		_CODE	0X443,"HAR",CHAR
		_CODE	0X554,"OKEN",TOKEN
		_CODE	0X457,"ORD",WORDD
		_CODE	0X54E,"AME>",NAMET
		_CODE	0X553,"AME?",SAMEQ
		_CODE	0X466,"ind",FIND
		_CODE	0X54E,"AME?",NAMEQ
		_CODE	0X25E,"H",BKSP
		_CODE	0X354,"AP",TAP
		_CODE	0X46B,"TAP",KTAP
		_CODE	0X661,"ccept",ACCEP
		_CODE	0X645,"XPECT",EXPEC
		_CODE	0X551,"UERY",QUERY
		_CODE	0X543,"ATCH",CATCH
		_CODE	0X554,"HROW",THROW
		_CODE	0X54E,"NULL$",NULLS
		_CODE	0X541,"BORT",ABORT
		_CODE	COMPO+0X661,"bort"",ABORQ
		_CODE	0X0A24,"INTERPRET",INTER
		_CODE	IMEDD+0X15B,"",LBRAC
		_CODE	0X32E,"OK",DOTOK
		_CODE	0X63F,"STACK",QSTAC
		_CODE	0X445,"VAL",EVAL
		_CODE	0X650,"RESET",PRESE
		_CODE	COMPO+0X378,"io",XIO
		_CODE	0X446,"ILE",FILE
		_CODE	0X448,"AND",HAND
		_CODE	0X349,"/O",ISLO
		_CODE	0X743,"ONSOLE",CONSO
		_CODE	0X451,"UIT",QUIT
		_CODE	0X127,"",TICK
		_CODE	0X541,"LLOT",ALLOT
		_CODE	0X12C,"",COMMA
		_CODE	IMEDD+0X95B,"COMPILE]",BCOMP
		_CODE	COMPO+0X743,"OMPILE",COMPI
		_CODE	IMEDD+0X74C,"ITERAL",LITER
		_CODE	0X324,0X2C22,STRCQ
		_CODE	IMEDD+0X752,"ECURSE",RECUR
		_CODE	IMEDD+0X346,"OR",FOR
		_CODE	IMEDD+0X542,"EGIN",BEGIN
		_CODE	IMEDD+0X44E,"EXT",NEXTT
		_CODE	IMEDD+0X555,"NTIL",UNTIL
		_CODE	IMEDD+0X541,"GAIN",AGAIN
		_CODE	IMEDD+0X249,"F",IFF
		_CODE	IMEDD+0X541,"HEAD",AHEAD
		_CODE	IMEDD+0X652,"EPEAT",REPEA
		_CODE	IMEDD+0X454,"HEN",THENN
		_CODE	IMEDD+0X341,"FT",AFT
		_CODE	IMEDD+0X445,"LSE",ELSEE
		_CODE	IMEDD+0X557,"HILE",WHILE
		_CODE	IMEDD+0X641,"BORT"",ABRTQ
		_CODE	IMEDD+0X224,0X2200,STRQ
		_CODE	IMEDD+0X22E,0X2200,DOTQ
		_CODE	0X73F,"UNIQUE",UNIQU
		_CODE	0X324,",n",SNAME
		_CODE	0X824,"COMPILE",SCOMP
		_CODE	0X54F,"VERT",OVERT
		_CODE	IMEDD+COMPO+0X13B,"",SEMIS
		_CODE	0X15D,"",RBRAC
		_CODE	0X563,"all,",CALLC
		_CODE	0X13A,"",COLON
		_CODE	0X949,"MMEDIATE",IMMED
		_CODE	0X843,"ONSTANT",CONST
		_CODE	0X455,"SER",USER
		_CODE	0X643,"REATE",CREAT
		_CODE	0X856,"ARIABLE",VARIA
		_CODE	0X55F,"TYPE",UTYPE
		_CODE	0X364,"mp",DMP
		_CODE	0X444,"UMP",DUMP
		_CODE	0X22E,"S",DOTS
		_CODE	0X421,"CSP",STCSP
		_CODE	0X43F,"CSP",QCSP
		_CODE	0X53E,"NAME",TNAME
		_CODE	0X32E,"ID",DOTID
		_CODE	0X353,"EE",SEE
		_CODE	0X557,"ORDS",WORDS
		_CODE	0X356,"ER",VERSN
		_CODE	0X268,"i",HI
		_CODE	0X527,"BOOT",TBOOT
		_CODE	0X443,"OLD",COLD

CTOP:		EQU	0X4200		    ; user code dictionary
NTOP:		EQU	0X7FF0		    ; user name dictionary


; Initial values for user variables

UZERO:	DW	0	    ; for multitasker
	DW	0
	DW	0
	DW	0
	DW	0x90	    ; SP0
	DW	0x20	    ; RP0
	DW	QRX	    ; '?KEY
	DW	TXSTO	    ; 'EMIT
	DW	ACCEP	    ; 'EXPECT
	DW	KTAP	    ; 'TAP
	DW	TXSTO	    ; 'ECHO
	DW	DOTOK	    ; 'PROMPT
	DW	BASEE	    ; BASE
	DW	0	    ; tmp
	DW	0	    ; SPAN
	DW	0	    ; >IN
	DW	0	    ; #TIB
	DW	TIB	    ; TIB
	DW	0	    ; CSP
	DW	INTER	    ; 'EVAL
	DW	NUMBQ	    ; 'NUMBER
	DW	0	    ; HLD
	DW	0	    ; HANDLER
	DW	0	    ; CONTEXT
	DW	0	    ; vocabulary stack
	DW	0
	DW	0
	DW	0
	DW	0
	DW	0
	DW	0
	DW	0
	DW	0	    ; CURRENT
	DW	0	    ; vocabulary link
	DW	CTOP	    ; CP
	DW	NTOP	    ; NP
	DW	LINK220     ; LAST, COLD
	DW	LINK220     ; FORTH vocabulary pointer
	DW	0	    ; EDITOR vocabulary pointer, as needed
	DW	0	    ; ASSEMBLER vocabulary pointer
	DW	0	    ; APPLICATION vocabulary pointer
ULAST:	DW	0	    ; dummy


; Cold start initiation of the virtual Forth engine

COLDD:	clrf	INTSTA,1
	movlw	0x50	    ; post-increment SP and RP
	movwf	ALUSTA
	movlw	1
	movwf	BSR	    ; select bank 1
	clrf	PIE	    ; disable all interrupts
	clrf	BSR	    ; select bank 0
	movlw	RPCPU	    ; return stack
	movwf	RP
	movlw	SPCPU
	movwf	SP
	movlw	UPP	    ; init UP pointer in 4000
	movwf	TBL
	movlw	HIGH UPP
	movwf	TBH
	movlw	HIGH USERA  ; store in it 4080, default UP
	tlwt	1,W
	movlw	USERA
	tlwt	0,W
	tablwt	0,0,W

	movlw	WARM1	    ; init IP
	movwf	TBL
	movlw	HIGH WARM1  ; prepare warm start
	movwf	TBH
	movfp	TBL,RPP     ; push IP on return stack for safety
	movfp	TBH,RPP
	NEXT


; Diagnostic routines to test all the low level code routines
; Lifted from Ken Chen's Z80 eForth implementation

;DIAG:	goto	DOLST
;eCHK:	DW	STOIO
;	DW	DOLIT		; test transmitter
;	DW	" e"
;	DW	TXSTO
;
;FCHK:	DW	DOLIT		; test um+, 0<
;	DW	0
;	DW	ZLESS
;	DW	DOLIT
;	DW	-2
;	DW	ZLESS
;	DW	UPLUS
;	DW	DROP		; -1 is expected here
;	DW	DOLIT
;	DW	3
;	DW	UPLUS
;	DW	UPLUS
;	DW	DROP
;	DW	DOLIT
;	DW	" C"		 ; "F"-3
;	DW	UPLUS
;	DW	DROP
;	DW	TXSTO		; print F
;
;oCHK:	DW	DOLIT		; test XOR, AND, OR
;	DW	" O"
;	DW	DOLIT
;	DW	" o"
;	DW	XORR
;	DW	DOLIT
;	DW	0xF0
;	DW	ANDD
;	DW	DOLIT
;	DW	" O"
;	DW	ORR
;	DW	TXSTO		; print o
;
;rCHK:	DW	SPAT		; test DUP SWAP DROP OVER SP@ SP!
;	DW	SPAT
;	DW	SWAP
;	DW	OVER
;	DW	XORR
;	DW	DOLIT
;	DW	3
;	DW	ANDD
;	DW	DOLIT
;	DW	" p"		 ; "r"-2
;	DW	UPLUS
;	DW	DROP
;	DW	SWAP
;	DW	DUPP
;	DW	DUPP
;	DW	SPSTO
;	DW	TXSTO		; print r
;
;tCHK:	DW	DOLIT		; test BRAN, QBRAN
;	DW	0
;	DW	QBRAN
;QB1:	DW	QB1+4		; QB2
;	DW	DOLIT
;	DW	" ?"
;	DW	TXSTO		; ? QBRAN error
;QB2:	DW	DOLIT
;	DW	-1
;	DW	QBRAN
;	DW	QB1+13		; QB3
;	DW	DOLIT
;	DW	" t"
;	DW	TXSTO		; print t if no error
;	DW	BRAN
;	DW	QB1+16		; hCHK
;QB3:	DW	DOLIT
;	DW	" !"
;	DW	TXSTO		; ! BRAN error
;
;hCHK:	DW	DOLIT		; test @ !
;	DW	" h"
;	DW	DOLIT
;	DW	RAM+1
;	DW	STORE
;	DW	DOLIT
;	DW	RAM+1
;	DW	AT
;	DW	DOLIT
;	DW	RAM*2+2
;	DW	CSTOR
;	DW	DOLIT
;	DW	RAM*2+2
;	DW	CAT
;	DW	TXSTO
;
;MCHK:	DW	DOLIT		; test >R R> R@ RP@ RP!
;	DW	" M"
;	DW	TOR
;	DW	RPAT
;	DW	DOLIT
;	DW	" ?"
;	DW	TOR
;	DW	RPSTO
;	DW	RFROM
;	DW	TXSTO
;
;UCHK:	DW	SZERO		; test DOUSE
;	DW	AT
;	DW	DOLIT
;	DW	0x90
;	DW	XORR
;	DW	QBRAN
;	DW	UCHK+10
;	DW	DOLIT
;	DW	" ?"
;	DW	TXSTO
;UCHK1:	DW	DOLIT
;	DW	" u"
;	DW	TXSTO
;
;	DW	DOLIT		; test FOR NEXT
;	DW	0
;	DW	DOLIT
;	DW	10
;	DW	TOR
;FOR1:	DW	DOLIT
;	DW	1
;	DW	UPLUS
;	DW	DROP
;	DW	DONXT
;	DW	FOR1
;	DW	DOLIT
;	DW	11
;	DW	XORR
;	DW	QBRAN
;UCHK2:	DW	UCHK2+5 	; NXTOK
;	DW	DOLIT
;	DW	" ?"
;	DW	TXSTO
;	DW	EXIT
;NXTOK:	DW	DOLIT
;	DW	" l"
;	DW	TXSTO
;	DW	EXIT
;

;;	warm start

WARM:  goto    DOLST
WARM1: DW      DOLIT	       ; init user area
       DW      UZERO
       DW      DOLIT
       DW      USERA
       DW      DOLIT
       DW      ULAST-UZERO
       DW      MOVE
       DW      COLD		 ; or to DIAG
       DW      WARM


	END
